// model.c

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <gl\gl.h>
#include <assert.h>

#include "model.h"
#include "tga.h"
#include "Matrix.h"

// Vector x Matrix (based on Brett Porter's Vector class)
void Vec_Transform( VEC * pV, const MATRIX * pM )
{
	double vector[4];
	const float * matrix = Matrix_GetMatrix(pM);

	vector[0] = pV->x * matrix[0] + pV->y * matrix[4] + pV->z * matrix[8] + matrix[12];
	vector[1] = pV->x * matrix[1] + pV->y * matrix[5] + pV->z * matrix[9] + matrix[13];
	vector[2] = pV->x * matrix[2] + pV->y * matrix[6] + pV->z * matrix[10] + matrix[14];
	vector[3] = pV->x * matrix[3] + pV->y * matrix[7] + pV->z * matrix[11] + matrix[15];

	pV->x = ( float )( vector[0] );
	pV->y = ( float )( vector[1] );
	pV->z = ( float )( vector[2] );
	pV->w = ( float )( vector[3] );
}

void Vec_Transform3( VEC * pV, const MATRIX * pM )
{
	double vector[3];
	const float *matrix = Matrix_GetMatrix(pM);

	vector[0] = pV->x * matrix[0] + pV->y * matrix[4] + pV->z * matrix[8];
	vector[1] = pV->x * matrix[1] + pV->y * matrix[5] + pV->z * matrix[9];
	vector[2] = pV->x * matrix[2] + pV->y * matrix[6] + pV->z * matrix[10];

	pV->x = ( float )( vector[0] );
	pV->y = ( float )( vector[1] );
	pV->z = ( float )( vector[2] );
}

// Mesh
void Mesh_Delete( MESH * mesh )
{
	if(mesh->vertices != NULL)
		free(mesh->vertices);

	if(mesh->triangles != NULL)
		free(mesh->triangles);

	if(mesh->normals != NULL)
		free(mesh->normals);
}

BOOL Mesh_LoadFromMs3dAsciiSegment( MESH * mesh, FILE *file )
{
	BOOL bError = FALSE;
	char szLine[256] = { 0 };
	int nFlags, nIndex, j;

	// vertices
	if (!fgets (szLine, 256, file))
	{
		return FALSE;
	}

	if (sscanf (szLine, "%d", &mesh->numVertices) != 1)
	{
		return FALSE;
	}
	mesh->vertices = malloc( sizeof(VEC) * mesh->numVertices);

	for (j = 0; j < mesh->numVertices; j++)
	{
		if (!fgets (szLine, 256, file))
		{
			return FALSE;
		}
		if (sscanf (szLine, "%d %f %f %f %f %f %d",
			&nFlags,
			&mesh->vertices[j].x, &mesh->vertices[j].y, &mesh->vertices[j].z,
			&mesh->vertices[j].u, &mesh->vertices[j].v,
			&mesh->vertices[j].bone) != 7){
			return FALSE;
		}

		// adjust the y direction of the texture coordinate
		mesh->vertices[j].v = 1.0f - mesh->vertices[j].v;
	}

	// normals
	if (!fgets (szLine, 256, file))
	{
		return FALSE;
	}

	if (sscanf (szLine, "%d", &mesh->numNormals) != 1)
	{
		return FALSE;
	}
	mesh->normals = malloc( sizeof(NORMAL) * mesh->numNormals);

	for (j = 0; j < mesh->numNormals; j++)
	{
		if (!fgets (szLine, 256, file))
		{
			return FALSE;
		}

		if (sscanf (szLine, "%f %f %f",
			&mesh->normals[j].x, &mesh->normals[j].y, &mesh->normals[j].z) != 3)
		{
			return FALSE;
		}
	}

	// triangles
	if (!fgets (szLine, 256, file))
	{
		return FALSE;
	}

	if (sscanf (szLine, "%d", &mesh->numTriangles) != 1)
	{
		return FALSE;
	}
	mesh->triangles = malloc( sizeof(TRI) * mesh->numTriangles);

	for (j = 0; j < mesh->numTriangles; j++)
	{
		if (!fgets (szLine, 256, file))
		{
			return FALSE;
		}

		if (sscanf (szLine, "%d %d %d %d %d %d %d %d",
			&nFlags,
			&mesh->triangles[j].v[0], &mesh->triangles[j].v[1], &mesh->triangles[j].v[2],
			&mesh->triangles[j].n[0], &mesh->triangles[j].n[1], &mesh->triangles[j].n[2],
			&nIndex) != 8){
			return FALSE;
		}
		assert(mesh->triangles[j].v[0] >= 0);
		assert(mesh->triangles[j].v[0] < mesh->numVertices);
		assert(mesh->triangles[j].v[1] >= 0);
		assert(mesh->triangles[j].v[1] < mesh->numVertices);
		assert(mesh->triangles[j].v[2] >= 0);
		assert(mesh->triangles[j].v[2] < mesh->numVertices);
	}

	return TRUE;
}

// Material
void Material_ReloadTexture( MATERIAL * material )
{
	if( strlen(material->diffuse_name) > 0 )
		TGA_LoadTexture(material->diffuse_name, &material->texture);
	else
	    material->texture = 0;
}

BOOL Material_LoadFromMs3dAsciiSegment(MATERIAL * material, char * path, FILE *file )
{
	char szLine[256] = { 0 };

	// name
	if (!fgets (szLine, 256, file))
		return FALSE;
	if (sscanf (szLine, "\"%[^\"]\"", material->name) != 1)
		return FALSE;

	// ambient
	if (!fgets (szLine, 256, file))
		return FALSE;

	if (sscanf (szLine, "%f %f %f %f", &material->ambient[0], &material->ambient[1],
		&material->ambient[2], &material->ambient[3]) != 4)
		return FALSE;

	// diffuse
	if (!fgets (szLine, 256, file))
		return FALSE;
	if (sscanf (szLine, "%f %f %f %f", &material->diffuse[0], &material->diffuse[1],
		&material->diffuse[2], &material->diffuse[3]) != 4)
		return FALSE;

	// specular
	if (!fgets (szLine, 256, file))
		return FALSE;
	if (sscanf (szLine, "%f %f %f %f", &material->specular[0], &material->specular[1],
		&material->specular[2], &material->specular[3]) != 4)
		return FALSE;

	// emissive
	if (!fgets (szLine, 256, file))
		return FALSE;
	if (sscanf (szLine, "%f %f %f %f", &material->emissive[0], &material->emissive[1],
		&material->emissive[2], &material->emissive[3]) != 4)
		return FALSE;

	// shininess
	if (!fgets (szLine, 256, file))
		return FALSE;
	if (sscanf (szLine, "%f", &material->shininess) != 1)
		return FALSE;

	// transparency
	if (!fgets (szLine, 256, file))
		return FALSE;
	if (sscanf (szLine, "%f", &material->transparency) != 1)
		return FALSE;

	// diffuse texture
	if (!fgets (szLine, 256, file))
		return FALSE;
	strcpy(material->diffuse_name, "");
	char name[MAX_PATH] = {	0 };
	sscanf(szLine, "\"%[^\"]\"", name);
	sprintf(material->diffuse_name, "%s%s", path, name);

	// alpha texture
	if (!fgets (szLine, 256, file))
		return FALSE;
	strcpy(material->alpha_name, "");
	sscanf (szLine, "\"%[^\"]\"", material->alpha_name);

	Material_ReloadTexture(material);

	return TRUE;
}

void Bone_Delete(BONE  * bone)
{
	if(bone->keyFramesTrans != NULL)
	{
		free(bone->keyFramesTrans);
		bone->keyFramesTrans = NULL;
	}

	if(bone->keyFramesRot != NULL)
	{
		free(bone->keyFramesRot);
		bone->keyFramesRot = NULL;
	}
}

void Bone_Render(BONE * bone)
{
	VEC	vector;
	VEC	parentvector;

	vector.x = 0.0f;
	vector.y = 0.0f;
	vector.z = 0.0f;
	vector.w = 1.0f;
	Vec_Transform(&vector, &bone->matFinal );

	if( bone->parent != NULL )
	{
		parentvector.x = 0.0f;
		parentvector.y = 0.0f;
		parentvector.z = 0.0f;
		parentvector.w = 1.0f;
		Vec_Transform(&parentvector, &bone->parent->matFinal );
	}

	glDisable( GL_TEXTURE_2D );
	//	glDisable(GL_LIGHTING);

	// render bone as a line
	glLineWidth(1.0f);
	glColor3f(1.0, 0.0f, 0.0f);
	glBegin(GL_LINES);
	glVertex3f( vector.x, vector.y, vector.z );

	if(bone->parent != NULL)
		glVertex3f( parentvector.x, parentvector.y, parentvector.z );
	else
	    glVertex3f( vector.x, vector.y, vector.z );

	glEnd();

	// render bone-ends as fat points
	glPointSize(4.0f);
	glColor3f(1.0f, 0.0f, 1.0f);
	glBegin(GL_POINTS);
	glVertex3f( vector.x, vector.y, vector.z );

	if(bone->parent != NULL)
		glVertex3f( parentvector.x, parentvector.y, parentvector.z );
	glEnd();

	glColor3f(1.0f, 1.0f, 1.0f);

}

BOOL Bone_LoadFromMs3dAsciiSegment(BONE * bone, FILE *file )
{
	char szLine[256] = { 0 };
	int	 j;
	int	 nFlags;

	// name
	if (!fgets (szLine, 256, file))
		return FALSE;

	if (sscanf (szLine, "\"%[^\"]\"", bone->name) != 1)
		return FALSE;

	// parent name
	if (!fgets (szLine, 256, file))
		return FALSE;
	strcpy (bone->parentName, "");
	sscanf (szLine, "\"%[^\"]\"", bone->parentName);


	if (!fgets (szLine, 256, file))
		return FALSE;
	if (sscanf (szLine, "%d %f %f %f %f %f %f",
		&nFlags,
		&bone->position[0], &bone->position[1], &bone->position[2],
		&bone->rotation[0], &bone->rotation[1], &bone->rotation[2]) != 7)
		return FALSE;


	// position key count
	if (!fgets (szLine, 256, file))
		return FALSE;
	if (sscanf (szLine, "%d", &bone->numPosKeys) != 1)
		return FALSE;

	bone->keyFramesTrans = malloc( sizeof(KEYFRAMEPOS) * bone->numPosKeys);
	for (j = 0; j < bone->numPosKeys; j++)
	{
		if (!fgets (szLine, 256, file))
			return FALSE;
		if (sscanf (szLine, "%f %f %f %f", &bone->keyFramesTrans[j].time,
			&bone->keyFramesTrans[j].position[0],
			&bone->keyFramesTrans[j].position[1],
			&bone->keyFramesTrans[j].position[2] ) != 4)
			return FALSE;
	}

	// rotation key count
	if (!fgets (szLine, 256, file))
		return FALSE;
	if (sscanf (szLine, "%d", &bone->numRotKeys) != 1)
		return FALSE;

	bone->keyFramesRot = malloc( sizeof(KEYFRAMEROT) * bone->numRotKeys);
	for (j = 0; j < bone->numRotKeys; j++)
	{
		if (!fgets (szLine, 256, file))
			return FALSE;
		if (sscanf (szLine, "%f %f %f %f", &bone->keyFramesRot[j].time,
			&bone->keyFramesRot[j].position[0],
			&bone->keyFramesRot[j].position[1],
			&bone->keyFramesRot[j].position[2] ) != 4)
			return FALSE;
	}

	return TRUE;
}

void Bone_Initialize(BONE * bone)
{
	MATRIX m_rel = { 0.0f };

	// Create a transformation matrix from the position and rotation

	Matrix_SetRotationRadians(&m_rel, bone->rotation );
	Matrix_SetTranslation(&m_rel, bone->position );

	// Each bone's final matrix is its relative matrix concatenated onto its
	// parent's final matrix (which in turn is ....)
	//
	if ( bone->parent == NULL ){
		Matrix_Set(&bone->matFinal, Matrix_GetMatrix(&m_rel));
	}else
	    {
		Matrix_Set(&bone->matFinal, Matrix_GetMatrix(&bone->parent->matFinal));
		Matrix_PostMultiply(&bone->matFinal, &m_rel );
	}
}

void Bone_AdvanceTo(BONE * bone, float fCurrentTime )
{
	int		i;
	float	deltaTime;
	float	fraction;
	float	Position[3];
	float	Rotation[3];

	/*
	* Position
	*/

	// Find appropriate position key frame
	i = 0;
	while( i < bone->numPosKeys - 1 && bone->keyFramesTrans[i].time < fCurrentTime )
		i++;

	assert(i < bone->numPosKeys);

	if( i > 0 )
	{
		// Interpolate between 2 key frames

		// time between the 2 key frames
		deltaTime = bone->keyFramesTrans[i].time - bone->keyFramesTrans[i-1].time;

		assert( deltaTime > 0 );

		// relative position of interpolation point to the keyframes [0..1]
		fraction = (fCurrentTime - bone->keyFramesTrans[i-1].time) / deltaTime;

		assert( fraction > 0 );
		assert( fraction < 1.0 );

		Position[0] = bone->keyFramesTrans[i-1].position[0] + fraction * (bone->keyFramesTrans[i].position[0] - bone->keyFramesTrans[i-1].position[0]);
		Position[1] = bone->keyFramesTrans[i-1].position[1] + fraction * (bone->keyFramesTrans[i].position[1] - bone->keyFramesTrans[i-1].position[1]);
		Position[2] = bone->keyFramesTrans[i-1].position[2] + fraction * (bone->keyFramesTrans[i].position[2] - bone->keyFramesTrans[i-1].position[2]);
	}
	else
	    {
		memcpy(Position, bone->keyFramesTrans[i].position, 3*sizeof(float) );
	}

	/*
	* Rotation
	*/

	// Find appropriate rotation key frame
	i = 0;
	while( i < bone->numRotKeys-1 && bone->keyFramesRot[i].time < fCurrentTime )
		i++;

	assert(i < bone->numRotKeys);

	if( i > 0 )
	{
		// Interpolate between 2 key frames

		// time between the 2 key frames
		deltaTime = bone->keyFramesRot[i].time - bone->keyFramesRot[i-1].time;
		assert( deltaTime > 0 );

		// relative position of interpolation point to the keyframes [0..1]
		fraction = (fCurrentTime - bone->keyFramesRot[i-1].time) / deltaTime;
		assert( fraction > 0 );
		assert( fraction < 1.0 );

		Rotation[0] = bone->keyFramesRot[i-1].position[0] + fraction * (bone->keyFramesRot[i].position[0] - bone->keyFramesRot[i-1].position[0]);
		Rotation[1] = bone->keyFramesRot[i-1].position[1] + fraction * (bone->keyFramesRot[i].position[1] - bone->keyFramesRot[i-1].position[1]);
		Rotation[2] = bone->keyFramesRot[i-1].position[2] + fraction * (bone->keyFramesRot[i].position[2] - bone->keyFramesRot[i-1].position[2]);
	}
	else
	    {
		memcpy(Rotation, bone->keyFramesRot[i].position, 3*sizeof(float) );
	}

	// Now we know the position and rotation for this animation frame.
	// Let's calculate the transformation matrix (matFinal) for this bone...

	MATRIX m_rel = {0.0f};
	MATRIX m_frame = {0.0f};

	// Create a transformation matrix from the position and rotation of this
	// joint in the rest position
	Matrix_SetRotationRadians(&m_rel, bone->rotation );
	Matrix_SetTranslation(&m_rel, bone->position );

	// Create a transformation matrix from the position and rotation
	// m_frame: additional transformation for this frame of the animation
	Matrix_SetRotationRadians(&m_frame, Rotation );
	Matrix_SetTranslation(&m_frame, Position );

	// Add the animation state to the rest position
	Matrix_PostMultiply(&m_rel, &m_frame );

	//
	if ( bone->parent == NULL )						// this is the root node

		Matrix_Set(&bone->matFinal, Matrix_GetMatrix(&m_rel));	// matFinal := m_rel
	else										// not the root node
	{
		// matFinal := parent's matFinal * m_rel (matrix concatenation)
		Matrix_Set(&bone->matFinal, Matrix_GetMatrix(&bone->parent->matFinal));
		Matrix_PostMultiply(&bone->matFinal, &m_rel );
	}
}

/*
* Model
*/
void Model_Model(MODEL * pMod)
{
	pMod->numMeshes    = 0;
	pMod->pMeshes 	   = NULL;
	pMod->numMaterials = 0;
	pMod->pMaterials   = NULL;
}

void Model_Delete(MODEL * pMod)
{
	int i;
	if(pMod->pMeshes != NULL)
	{
		Mesh_Delete( pMod->pMeshes );
		free(pMod->pMeshes);
		pMod->pMeshes = NULL;
	}

	if(pMod->pMaterials != NULL)
	{
		free(pMod->pMaterials);
		pMod->pMaterials = NULL;
	}
	if(pMod->pBones != NULL){
		Bone_Delete( pMod->pBones );
		free(pMod->pBones);
		pMod->pBones = NULL;
	}

	if(pMod->materialIndices != NULL)
	{
		free(pMod->materialIndices);
		pMod->materialIndices = NULL;
	}
}

BOOL Model_LinkBones(MODEL * pMod)
{
	// The relationship between child and parent bones are defined
	// by each child bone providing the name of its parent bone.
	// This function builds corresponding pointers for faster lookup.
	//
	int i, j;

	// Link children to parent
	//

	for (i = 0; i < pMod->numBones; i++)
	{
		pMod->pBones[i].parent = NULL;

		if( strlen(pMod->pBones[i].parentName) >  0 )	// does bone have parent?
		{
			for (j = 0; j < pMod->numBones; j++)	// search for parent
			{
				if( stricmp(pMod->pBones[j].name, pMod->pBones[i].parentName) == 0)	// j is parent of i
				{
					pMod->pBones[i].parent = &pMod->pBones[j];
					break;
				}
			}

			if ( pMod->pBones[i].parent == NULL) // Unable to find parent bone
				return FALSE;
		}
	}

	return TRUE;
}

void Model_InitializeBones(MODEL * pMod)
{
	int i;

	for (i = 0; i < pMod->numBones; i++)
	{
		Bone_Initialize(&pMod->pBones[i]);
	}
}

// TODO: Uses private members from various classes (temporarily made public). Clean up!
//
void Model_AttachSkin(MODEL * pMod)
{
	int		i, j;
	int		bone;
	MATRIX	* pMatrix;
	float	v[3];

	// Make vertex vector relative to the bone it's attached to

	for( i = 0; i < pMod->numMeshes; i++)	// for all meshes
	{
		for(j = 0; j < pMod->pMeshes[i].numVertices; j++ )	// for all vertices
		{
			bone = pMod->pMeshes[i].vertices[j].bone;

			if( bone != -1 )	// if vertex attached to a bone

			{

				pMatrix = &pMod->pBones[bone].matFinal;

				// make relative to bone position and orientation
				// modifies vertex x,y,z

				v[0] = pMod->pMeshes[i].vertices[j].x;
				v[1] = pMod->pMeshes[i].vertices[j].y;
				v[2] = pMod->pMeshes[i].vertices[j].z;
				Matrix_InverseTranslateVect(pMatrix, v );
				Matrix_InverseRotateVect(pMatrix, v );
				pMod->pMeshes[i].vertices[j].x = v[0];
				pMod->pMeshes[i].vertices[j].y = v[1];
				pMod->pMeshes[i].vertices[j].z = v[2];
			}

		}
	}
}


void Model_RenderBones(MODEL * pMod)
{
	int	i;

	for(i = 0; i < pMod->numBones; i++)	// for each bone
	{
		Bone_Render(&pMod->pBones[i]);
	}
}

void Model_AdvanceAnimation(MODEL * pMod, float deltaTime )
{
	int	i;

	pMod->fCurrentTime += deltaTime;
	if( pMod->fCurrentTime > pMod->fMaxTime )	// this is a looped animation!

		pMod->fCurrentTime = 1.0;		// N.B. time starts at 1, not at 0!

	for (i = 0; i < pMod->numBones; i++)
	{
		Bone_AdvanceTo(&pMod->pBones[i], pMod->fCurrentTime );
	}
}

// For this example sowtex3.tga is in the data folder
// so we have to adjust the path. See inside the .txt file
// for the name of the texture to load.
BOOL Model_LoadFromMs3dAsciiFile(MODEL * pMod, const char *path, const char *filename )
{
	float	StartTime;
	BOOL	bError = FALSE;
	char	szpath[256];
	char	szLine[256] = {	0 };
	int		nFlags, nIndex, i;

	sprintf(szpath, "%s%s", path, filename);

	FILE *file1 = fopen ("data/hog.ms3d", "rb");

	FILE *file = fopen (szpath, "r");
	if (!file)
		return FALSE;

	pMod->fCurrentTime = 0;
	int nFrame;

	while (fgets (szLine, 256, file) != NULL  && !bError)
	{
		if (!strncmp (szLine, "//", 2))
			continue;

		if (sscanf (szLine, "Frames: %d", &nFrame) == 1)
		{
			pMod->fMaxTime = 1.0f * nFrame;		// length of animation
			continue;
		}
		if (sscanf (szLine, "Frame: %d", &nFrame) == 1)
		{
			StartTime = 1.0f * nFrame;
			continue;
		}

		// meshes
		if (sscanf (szLine, "Meshes: %d", &pMod->numMeshes) == 1)
		{
			char	szName[MS_MAX_NAME];

			pMod->pMeshes = malloc( sizeof(MESH) * pMod->numMeshes);
			pMod->materialIndices = malloc( sizeof(int) * pMod->numMeshes);

			for (i = 0; i < pMod->numMeshes && !bError; i++)
			{

				if (!fgets (szLine, 256, file))
				{
					bError = TRUE;
					break;
				}

				// mesh: name, flags, material index
				if (sscanf (szLine, "\"%[^\"]\" %d %d",szName, &nFlags, &nIndex) != 3)
				{
					bError = TRUE;
					MessageBeep(16);
					break;
				}
				pMod->materialIndices[i] = nIndex;

				if(!Mesh_LoadFromMs3dAsciiSegment(&pMod->pMeshes[i], file) )
				{
					bError = TRUE;
					MessageBeep(16);
					break;
				}
			}
			continue;
		}

		// materials
		if (sscanf (szLine, "Materials: %d", &pMod->numMaterials) == 1)
		{
			int i;

			pMod->pMaterials = malloc( sizeof(MATERIAL) * pMod->numMaterials);

			for (i = 0; i < pMod->numMaterials && !bError; i++)
			{
				if( !Material_LoadFromMs3dAsciiSegment(&pMod->pMaterials[i], path, file) )
				{
					MessageBeep(16);
					bError = TRUE;
					break;
				}
			}
			continue;
		}

		// bones
		if (sscanf (szLine, "Bones: %d", &pMod->numBones) == 1)
		{
			int i = 0;

			pMod->pBones = malloc( sizeof(BONE) * pMod->numBones);

			for (i = 0; i < pMod->numBones && !bError; i++)
			{
				if( !Bone_LoadFromMs3dAsciiSegment(&pMod->pBones[i], file) )
				{
					MessageBeep(16);
					bError = TRUE;
					break;
				}
			}
			continue;
		}
	}

	fclose (file);

	if( !Model_LinkBones(pMod) )
		return FALSE;
	Model_InitializeBones(pMod);
	Model_AttachSkin(pMod);
	Model_AdvanceAnimation(pMod, StartTime );
	return TRUE;
}



